
/******************************************************************************
* MODULE     : machine.gen.cc
* DESCRIPTION: generation of fast arithmetic machine code
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <string.gen.h>

#module code_machine
#import string

/******************************************************************************
* The address class
******************************************************************************/

int address_nr;
int temp_space;

struct address_rep: public concrete_struct {
  int reg_nr;
  int offset;
  inline address_rep (int reg_nr2, int offset2):
    reg_nr (reg_nr2), offset (offset2) {}
};

class address {
#import concrete (address, address_rep)
public:
  inline address (int reg_nr, int offset= 0):
    rep (new address_rep (reg_nr, offset)) {}
  operator string ();
};
#import code_concrete (address, address_rep)
inline address operator + (address addr, int plus) {
  return address (addr->reg_nr, addr->offset+ plus); }

/******************************************************************************
* The variable class
******************************************************************************/

int variable_nr= 0;

struct variable {
  int nr;
  inline variable (): nr (variable_nr++) {}
  inline variable (int nr2): nr (nr2) {}
  operator string ();
};

/******************************************************************************
* The label class
******************************************************************************/

int label_nr= 0;

struct label {
  int nr;
  inline label (): nr (label_nr++) {}
  operator string ();
};

/******************************************************************************
* Code generation
******************************************************************************/

string buffer;

void
generate (string s) {
  buffer << s;
}

#include "X386/x386.gen.cc"

#define argument(name,n)                       \
  address name(address_nr);                    \
  new_argument(n);
#define parameter(name)                        \
  new_parameter();                             \
  variable name;
#define temporary(name,n)                      \
  address name(address_nr,temp_space);         \
  new_temporary(n);

/******************************************************************************
* Basic stuff
******************************************************************************/

void
mov (address addr, int what, int n) {
  int i;
  for (i=0; i<n; i++)
    mov (addr+ i, what);
}

void
mov (address addr1, address addr2, int n) {
  int i;
  for (i=0; i<n; i++)
    mov (addr1+ i, addr2+ i);
}

/******************************************************************************
* Addition and substraction
******************************************************************************/

void
add2 (address addr1, address addr2, int n) {
  int i;
  add (addr1, addr2);
  for (i=1; i<n; i++)
    adc (addr1+ i, addr2+ i);
}

void
adc2 (address addr1, address addr2, int n) {
  int i;
  for (i=0; i<n; i++)
    adc (addr1+ i, addr2+ i);
}

void
add3 (address addr1, address addr2, address addr3, int n) {
  int i;
  add (addr1, addr2, addr3);
  for (i=1; i<n; i++)
    adc (addr1+ i, addr2+ i, addr3+ i);
}

void
adc3 (address addr1, address addr2, address addr3, int n) {
  int i;
  for (i=0; i<n; i++)
    adc (addr1+ i, addr2+ i, addr3+ i);
}

void
inc (address addr1, int n) {
  label l;
  int i;
  for (i=0; i<n; i++) {
    add (1, addr1+ i);
    if (i<n-1) jcc (l);
  }
  generate (l);
}

void
icc (address addr1, int n) {
  label l;
  int i;
  for (i=0; i<n; i++) {
    adc (0, addr1+ i);
    if (i<n-1) jcc (l);
  }
  generate (l);
}

void
sub2 (address addr1, address addr2, int n) {
  int i;
  sub (addr1, addr2);
  for (i=1; i<n; i++)
    sbb (addr1+ i, addr2+ i);
}

void
sbb2 (address addr1, address addr2, int n) {
  int i;
  for (i=0; i<n; i++)
    sbb (addr1+ i, addr2+ i);
}

void
sub3 (address addr1, address addr2, address addr3, int n) {
  int i;
  sub (addr1, addr2, addr3);
  for (i=1; i<n; i++)
    sbb (addr1+ i, addr2+ i, addr3+ i);
}

void
sbb3 (address addr1, address addr2, address addr3, int n) {
  int i;
  for (i=0; i<n; i++)
    sbb (addr1+ i, addr2+ i, addr3+ i);
}

void
dec (address addr1, int n) {
  label l;
  int i;
  for (i=0; i<n; i++) {
    sub (1, addr1+ i);
    if (i<n-1) jcc (l);
  }
  generate (l);
}

void
dcc (address addr1, int n) {
  label l;
  int i;
  for (i=0; i<n; i++) {
    sbb (0, addr1+ i);
    if (i<n-1) jcc (l);
  }
  generate (l);
}

/******************************************************************************
* Multiplication
******************************************************************************/

void
naive_mul (address addr1, address addr2, address addr3, int n) {
  int i,j,k;
  mov (addr1+2, 0, 2*n-2);
  for (k=0; k<=(2*n-2); k++)
    for (i= ((k-n+1) >? 0); i< (n <? (k+1)); i++) {
      j= k-i;
      if ((i==0) && (j==0)) mul (addr1, addr2, addr3);
      else {
	mul_add (addr1+ k, addr2+ i, addr3+ j);
	if (k<(2*n-2)) adc (0, addr1+k+1);
      }
    }
}

/******************************************************************************
* Code generation
******************************************************************************/

void
generate_add2 (int n) {
  routine ("add2_" * as_string (n));
  argument (Dest, n);
  argument (Arg, n);
  add2 (Dest, Arg, n);
  end_routine ();
}

void
generate_add3 (int n) {
  routine ("add3_" * as_string (n));
  argument (Dest, n);
  argument (Arg1, n);
  argument (Arg2, n);
  add3 (Dest, Arg1, Arg2, n);
  end_routine ();
}

void
generate_inc (int n) {
  routine ("inc_" * as_string (n));
  argument (Dest, n);
  inc (Dest, n);
  end_routine ();
}

void
generate_sub2 (int n) {
  routine ("sub2_" * as_string (n));
  argument (Dest, n);
  argument (Arg, n);
  sub2 (Dest, Arg, n);
  end_routine ();
}

void
generate_sub3 (int n) {
  routine ("sub3_" * as_string (n));
  argument (Dest, n);
  argument (Arg1, n);
  argument (Arg2, n);
  sub3 (Dest, Arg1, Arg2, n);
  end_routine ();
}

void
generate_dec (int n) {
  routine ("dec_" * as_string (n));
  argument (Dest, n);
  dec (Dest, n);
  end_routine ();
}

void
generate_mul (int n) {
  routine ("mul_" * as_string (n));
  argument (Dest, n);
  argument (Arg1, n);
  argument (Arg2, n);
  naive_mul (Dest, Arg1, Arg2, n);
  end_routine ();
}

void
generate_fft (int n) {
  routine ("fft_" * as_string (n));
  argument (Arg1, n+1);
  argument (Arg2, n+1);
  parameter (Shift); (void) Shift;
  temporary (Temp, n+1);

  label lb1;
  sub3 (Arg2, Arg1, Temp, n+1);
  jcc (lb1);
  inc (Arg2, n);
  mov (0, Arg2+ n);
  adc (0, Arg2+ n);
  generate (lb1);

  label lc1, lc2;
  add2 (Arg1, Temp, n+1);
  sub (Arg1, Arg1+ n);
  dcc (Arg1+ 1, n-1);
  jcc (lc1);
  inc (Arg2, n);
  mov (0, Arg2+ n);
  adc (0, Arg2+ n);
  jmp (lc2);
  generate (lc1);
  mov (0, Arg1+ n);
  generate (lc2);

  end_routine ();
}

/******************************************************************************
* Main program
******************************************************************************/

int
main () {
  /*
  generate_add2 (8);
  generate_add3 (8);
  generate_inc  (8);
  generate_sub2 (8);
  generate_sub3 (8);
  generate_dec  (8);
  generate_fft  (8);
  */
  generate_mul (8);
  cout << buffer << "\n";
  return 0;
}

#endmodule // code_machine
